

function createEnvironment(globeRadius)
    --set up scene
    scene = craft.scene()
    scene.sky.material.sky = color(72, 235, 215)       
    scene.sky.material.horizon = color(34, 30, 66)       
    scene.sky.material.ground = color(33, 70, 120)

    sphereMesh = PseudoMesh()   
    sphereMesh:addSphere({
        --            axis=vec3(0,1,0),
        size=3,
        centre=vec3(0,0,0)
    })
    
    sphereModel= sphereMesh:toModel()   
    
    mySphere=scene:entity()
    mySphere.model=sphereModel
    mySphere.material= craft.material(asset.builtin.Materials.Standard)
    mySphere.material.map = readImage(asset.builtin.Surfaces.Desert_Cliff_Roughness)
    mySphere.position = vec3(5,5,5)
    
    cubesAtVertices(mySphere, color(236, 67, 202))

    --make an entity with the the .obj icosahedron model
    globe3D = createIcosahedron(globeRadius)

    --generate data to split the icosahedron into smaller triangles
    positions = globe3D.model.positions
    indices = globe3D.model.indices
    local newPos, newInd = subdivide(positions, indices, globe3D)
   -- newPos, newInd = subdivide(newPos, newInd, globe3D)
    
    --use that data to make a new model
    newModel = craft.model()
    newModel.positions = newPos 
    newModel.indices = newInd
    newModel.uvs = generateUVs(newPos)
    newModel.normals = generateNormals(newPos)
    globe3D.model = newModel
    
    --make little cubes at the vertices for visualization
    cubesAtVertices(globe3D, color(236, 67, 202))
    
    return globe3D
end

function generateUVs(positions)
    local uvs = {}
    for _, pos in ipairs(positions) do
        local normalizedPos = pos:normalize()
        
        -- Compute theta and phi
        local theta = math.atan(normalizedPos.z, normalizedPos.x)
        local phi = math.acos(normalizedPos.y)
        
        -- Convert to UV coordinates
        local u = (theta + math.pi) / (2 * math.pi)
        local v = phi / math.pi
        
        table.insert(uvs, vec2(u, v))
    end
    return uvs
end

function generateNormals(positions)
    local normals = {}
    for _, pos in ipairs(positions) do
        local normal = pos:normalize()
        table.insert(normals, normal)
    end
    return normals
end


function cubesAtVertices(entity, aColor)
    -- Create a material for the cubes
    local mat = craft.material("Materials:Specular")
    mat.diffuse = aColor or color(64, 0, 255)
    
    local scale = entity.scale
    
    -- Iterate over the model's positions and place a cube at each position
    for _, pos in ipairs(entity.model.positions) do
        local worldPos = vec3(pos.x * scale.x, pos.y * scale.y, pos.z * scale.z)
        local e = scene:entity()
        e.position = worldPos + entity.position
        e.model = craft.model.cube(vec3(0.035, 0.035, 0.035))
        e.material = mat
    end
end

function subdivide(positions, indices, entity)
    local newPositions = {}
    local newIndices = {}
    local midpointCache = {}
    
    local function getMidpoint(p1, p2)
        local key = math.min(p1, p2) .. "_" .. math.max(p1, p2)
        if midpointCache[key] then
            return midpointCache[key]
        end
        
        local midpoint = (positions[p1] + positions[p2]) * 0.5
        
        -- Calculate the distance of the original position from the center
        local originalDistance = positions[p1]:len()
        
        -- Normalize the midpoint and scale it to the original distance
        midpoint = midpoint:normalize() * originalDistance
        
        table.insert(newPositions, midpoint)
        local index = #newPositions + #positions -- Adjusted index to account for original positions
        
        midpointCache[key] = index
        return index
    end
    
    for i = 1, #indices, 3 do
        local a, b, c = indices[i], indices[i+1], indices[i+2]
        
        -- Subdivide the triangle
        local ab = getMidpoint(a, b)
        local bc = getMidpoint(b, c)
        local ca = getMidpoint(c, a)
        
        -- Triangle 1
        table.insert(newIndices, a)
        table.insert(newIndices, ca)
        table.insert(newIndices, ab)
        
        -- Triangle 2
        table.insert(newIndices, b)
        table.insert(newIndices, ab)
        table.insert(newIndices, bc)
        
        -- Triangle 3
        table.insert(newIndices, c)
        table.insert(newIndices, bc)
        table.insert(newIndices, ca)
        
        -- Triangle 4 (central triangle)
        table.insert(newIndices, ab)
        table.insert(newIndices, ca)
        table.insert(newIndices, bc)
        
    end
    
    for i, v in ipairs(positions) do
        table.insert(newPositions, v)
    end
    
    return newPositions, newIndices
end

function createIcosahedron(radius)
    local icosahedron = scene:entity()
    icosahedron.model = craft.model(asset.icosahedron)
    icosahedron.scale = vec3(1,1,1) * radius -- Adjust the size of the globe
    icosahedron.material = craft.material(asset.builtin.Materials.Standard)
    icosahedron.material.map = readImage(asset.builtin.Surfaces.Desert_Cliff_Roughness)
    return icosahedron
end

function generateUVs(positions)
    local uvTab = {}
    local maxY = -math.huge
    local minY = math.huge
    
    -- Find the top and bottom Y-coordinates
    for _, pos in ipairs(positions) do
        maxY = math.max(maxY, pos.y)
        minY = math.min(minY, pos.y)
    end
    
    for _, pos in ipairs(positions) do
        if pos.y == maxY then
            -- Top vertex
            table.insert(uvTab, vec2(0.5, 1))
        elseif pos.y == minY then
            -- Bottom vertex
            table.insert(uvTab, vec2(0.5, 0))
        else
            -- Middle vertices
            local u = 0.5 + math.atan(pos.z, pos.x) / (2 * math.pi)
            local v = 0.5 - pos.y / maxY
            table.insert(uvTab, vec2(u, v))
        end
    end
    
    return uvTab
end